from abc import ABC
from functools import cached_property
from typing import Generic, Type, TypeVar

from typecasts import Typecasts

from platonic.queue.types import InternalType
from platonic.type_args import generic_type_args

ValueType = TypeVar('ValueType')


class BaseQueue(ABC, Generic[ValueType]):
    """Base class for queues."""

    internal_type: Type[InternalType]
    typecasts: Typecasts

    @cached_property
    def value_type(self) -> Type[ValueType]:
        """Extract the type of queue message from type args."""
        return generic_type_args(self)[0]



from typing import Callable, TypeVar

ValueType = TypeVar('ValueType')


def const(anything: ValueType) -> Callable[[], ValueType]:
    """Generate a function which always returns the same value."""
    return lambda: anything



from abc import abstractmethod
from functools import cached_property
from typing import Callable, Iterable

from platonic.queue.base import BaseQueue
from platonic.queue.message import Message
from platonic.queue.types import InternalType, ValueType


class InputQueue(Iterable[Message[ValueType]], BaseQueue[ValueType]):
    """Queue to read stuff from."""

    @cached_property
    def deserialize_value(self) -> Callable[[InternalType], ValueType]:
        """Deserialize a queue item from internal representation."""
        return self.typecasts[  # pragma: no cover
            self.internal_type,
            self.value_type,
        ]

    @abstractmethod
    def receive(self) -> Message[ValueType]:
        """Get next message from queue, without deleting it."""

    @abstractmethod
    def receive_with_timeout(self, timeout: int) -> Message[ValueType]:
        """
        For the given period of time, wait for a new message from the queue.

        Return the message if received, otherwise throw a ReceiveTimeout error.
        """

    @abstractmethod
    def acknowledge(self, message: Message[ValueType]) -> Message[ValueType]:
        """
        Indicate that the given message is correctly processed.

        By semantics, is equivalent to `queue.Queue.task_done()`.
        """

    @abstractmethod
    def acknowledgement(self, message: Message[ValueType]):
        """
        Acknowledgement context manager.

        Into this context manager, you can wrap any operation with a given
        Message. The context manager will automatically acknowledge the message
        when and if the code in its context completes successfully.
        """



"""
Key-value Mapping.

This module is a draft.
"""

from abc import ABC
from functools import cached_property
from typing import Callable, Mapping, Type, TypeVar

from typecasts import Typecasts

from platonic.type_args import generic_type_args

KeyType = TypeVar('KeyType')
ValueType = TypeVar('ValueType')
InternalType = TypeVar('InternalType')


class PlatonicMapping(
    Mapping[KeyType, ValueType],
    ABC,
):
    """Abstract Mapping definition."""

    internal_type: type
    typecasts: Typecasts

    @cached_property
    def key_type(self) -> Type[KeyType]:
        """Type of the keys."""
        return generic_type_args(self)[0]

    @cached_property
    def value_type(self) -> Type[ValueType]:
        """Type of the values."""
        return generic_type_args(self)[1]

    @cached_property
    def serialize_key(self) -> Callable[[KeyType], InternalType]:
        """Convert key into internal representation."""
        return self.typecasts[self.key_type, self.internal_type]

    @cached_property
    def deserialize_key(self) -> Callable[[InternalType], KeyType]:
        """Restore key from internal representation."""
        return self.typecasts[self.internal_type, self.key_type]

    @cached_property
    def serialize_value(self) -> Callable[[ValueType], InternalType]:
        """Convert value into internal representation."""
        return self.typecasts[self.value_type, self.internal_type]

    @cached_property
    def deserialize_value(self) -> Callable[[InternalType], ValueType]:
        """Restore value from internal representation."""
        return self.typecasts[self.internal_type, self.value_type]



import dataclasses
from typing import Generic, TypeVar

MessageValueType = TypeVar('MessageValueType')


@dataclasses.dataclass
class Message(Generic[MessageValueType]):
    """Queue message."""

    value: MessageValueType  # noqa: WPS110



from abc import abstractmethod
from functools import cached_property
from typing import Callable, Iterable

from platonic.queue.base import BaseQueue
from platonic.queue.message import Message
from platonic.queue.types import InternalType, ValueType


class OutputQueue(BaseQueue[ValueType]):
    """Queue to write stuff into."""

    @cached_property
    def serialize_value(self) -> Callable[[ValueType], InternalType]:
        """Serialize a queue item into internal representation."""
        return self.typecasts[  # pragma: no cover
            self.value_type,
            self.internal_type,
        ]

    @abstractmethod
    def send(self, instance: ValueType) -> Message[ValueType]:
        """
        Push a message into the queue.

        See `InputQueue.get()` about naming controversy.
        """

    def send_many(self, iterable: Iterable[ValueType]) -> None:
        """Put multiple messages into the queue."""
        for instance in iterable:  # pragma: no cover
            self.send(instance)



from dataclasses import dataclass
from typing import Any, Iterator, Tuple, get_args

from documented import DocumentedError


@dataclass
class TypeArgsError(DocumentedError):   # type: ignore
    """
    Cannot ascertain type arguments from intance or class.

    Object: {self.instance}
    """

    instance: Any   # type: ignore


def iterate_by_type_parameters(  # type: ignore
    instance: Any,
) -> Iterator[Tuple[type, ...]]:
    """Iterate over classes in MRO and yield type parameters for each."""
    try:
        yield get_args(instance.__orig_class__)   # noqa: WPS609
    except AttributeError:
        pass

    try:
        yield from map(
            get_args,
            instance.__orig_bases__,  # noqa: WPS609
        )
    except AttributeError:
        pass

    if instance_args := get_args(instance):
        yield instance_args


def generic_type_args(instance: Any) -> Tuple[type, ...]:  # type: ignore
    """Get type parameters given a class instance."""
    try:
        type_parameters, *etc = iterate_by_type_parameters(instance)
    except ValueError:
        raise TypeArgsError(instance=instance)

    return type_parameters



import dataclasses

from documented import DocumentedError

from platonic.queue.base import BaseQueue
from platonic.queue.input import InputQueue, Message


@dataclasses.dataclass
class MessageReceiveTimeout(DocumentedError):
    """
    No messages received within {self.timeout} {self.readable_time_suffix}.

      Queue: {self.queue}
    """

    queue: InputQueue
    timeout: int

    @property
    def readable_time_suffix(self) -> str:
        """Second or seconds."""
        if self.timeout == 1:
            return 'second'

        return 'seconds'


@dataclasses.dataclass
class QueueDoesNotExist(DocumentedError):
    """Specified queue does not exist."""

    queue: BaseQueue


@dataclasses.dataclass
class MessageDoesNotExist(DocumentedError):
    """Specified message {self.message} does not exist in the queue."""

    message: Message
    queue: InputQueue


@dataclasses.dataclass
class MessageTooLarge(DocumentedError):
    """
    Message is too large.

    Provided message size is {self.message_size} bytes, while
    maximum message size is {self.max_supported_size} bytes.

    Message preview:
        {self.message_head}...{self.message_tail}
    """

    max_supported_size: int
    message_body: str

    message_preview_size: int = 10

    @property
    def message_head(self):
        """First N bytes of the message."""
        return self.message_body[:self.message_preview_size]

    @property
    def message_tail(self):
        """Last N bytes of the message."""
        return self.message_body[-self.message_preview_size:]

    @property
    def message_size(self) -> int:
        """Size of the raw message representation, in bytes."""
        return len(self.message_body)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['platonic', 'platonic.queue']

package_data = \
{'': ['*']}

install_requires = \
['typecasts>=0.1.1,<0.2.0']

setup_kwargs = {
    'name': 'platonic',
    'version': '0.1.2',
    'description': 'Abstract datastructures for Clean Architecture applications in Python.',
    'long_description': '# platonic\n\n[![Build Status](https://travis-ci.com/python-platonic/platonic.svg?branch=master)](https://travis-ci.com/python-platonic/platonic)\n[![Coverage](https://coveralls.io/repos/github/python-platonic/platonic/badge.svg?branch=master)](https://coveralls.io/github/python-platonic/platonic?branch=master)\n[![Python Version](https://img.shields.io/pypi/pyversions/platonic.svg)](https://pypi.org/project/platonic/)\n[![wemake-python-styleguide](https://img.shields.io/badge/style-wemake-000000.svg)](https://github.com/wemake-services/wemake-python-styleguide)\n![PyPI - License](https://img.shields.io/pypi/l/platonic)\n\nAbstract data structures for Clean Architecture applications in Python. Amazon services, brokers, and backends represented as queues, mappings, lists, iterables, and more.\n\n## Example \n\n```python\n# TODO\n```\n\n\n# Available data structures\n\n|                 | queue | iterable | dict | list | set | graph |\n| ---             | ---   | ---      | ---  | ---  | --- | ---   |\n| Amazon DynamoDB | 🔧    | 🔧       | 🔧   | 🔧   | 🔧  | 🔧    |\n| Amazon SimpleDB | 🔧    | 🔧       | 🔧   | 🔧   | 🔧  | 🔧    |\n| Amazon SQS      | ✔    ️| 🔧       | ❌    | 🔧   | 🔧  | 🔧    |\n| Amazon S3       | 🔧    | 🔧       | 🔧   | 🔧   | 🔧  | 🔧    |\n| Apache Kafka    | 🔧    | 🔧       | 🔧   | 🔧   | 🔧  | 🔧    |\n| Local FS        | 🔧    | 🔧       | 🔧   | 🔧   | 🔧  | 🔧    |\n| MongoDB         | 🔧    | 🔧       | 🔧   | 🔧   | 🔧  | 🔧    |\n| MySQL           | 🔧    | 🔧       | 🔧   | 🔧   | 🔧  | 🔧    |\n| OrientDB        | 🔧    | 🔧       | 🔧   | 🔧   | 🔧  | 🔧    |\n| PostgreSQL      | 🔧    | 🔧       | 🔧   | 🔧   | 🔧  | 🔧    |\n| Redis           | 🔧    | 🔧       | 🔧   | 🔧   | 🔧  | 🔧    |\n\n\n## Installation\n\n```bash\npip install platonic\n```\n\n## Credits\n\nThis project was generated with [`wemake-python-package`](https://github.com/wemake-services/wemake-python-package). Current template version is: [3cb37d9e138d83958c4d915a3b3aa737b27b6418](https://github.com/wemake-services/wemake-python-package/tree/3cb37d9e138d83958c4d915a3b3aa737b27b6418). See what is [updated](https://github.com/wemake-services/wemake-python-package/compare/3cb37d9e138d83958c4d915a3b3aa737b27b6418...master) since then.\n',
    'author': None,
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/python-platonic/platonic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from platonic.queue.errors import (
    MessageDoesNotExist,
    MessageReceiveTimeout,
    MessageTooLarge,
    QueueDoesNotExist,
)
from platonic.queue.input import InputQueue
from platonic.queue.message import Message
from platonic.queue.output import OutputQueue
from platonic.queue.types import InternalType, ValueType
